from unittest import mock

import pytest
from django.db import models
from rest_framework import serializers, viewsets
from rest_framework.decorators import api_view
from rest_framework.response import Response

from drf_spectacular.openapi import AutoSchema
from drf_spectacular.utils import (
    OpenApiParameter, PolymorphicProxySerializer, extend_schema, extend_schema_field,
)
from tests import assert_schema, generate_schema, get_request_schema, get_response_schema


class LegalPerson2(models.Model):
    company_name = models.CharField(max_length=30)


class NaturalPerson2(models.Model):
    first_name = models.CharField(max_length=30)
    last_name = models.CharField(max_length=30)


class LegalPersonSerializer(serializers.ModelSerializer):
    type = serializers.SerializerMethodField()

    class Meta:
        model = LegalPerson2
        fields = ('id', 'company_name', 'type')

    def get_type(self, obj) -> str:
        return 'legal'


class NaturalPersonSerializer(serializers.ModelSerializer):
    type = serializers.SerializerMethodField()

    class Meta:
        model = NaturalPerson2
        fields = ('id', 'first_name', 'last_name', 'type')

    def get_type(self, obj) -> str:
        return 'natural'


PROXY_SERIALIZER_PARAMS = {
    'component_name': 'MetaPerson',
    'serializers': [LegalPersonSerializer, NaturalPersonSerializer],
    'resource_type_field_name': 'type',
}


with mock.patch('rest_framework.settings.api_settings.DEFAULT_SCHEMA_CLASS', AutoSchema):
    implicit_poly_proxy = PolymorphicProxySerializer(
        component_name='MetaPerson',
        serializers=[LegalPersonSerializer, NaturalPersonSerializer],
        resource_type_field_name='type',
    )

    class ImplicitPersonViewSet(viewsets.GenericViewSet):
        @extend_schema(request=implicit_poly_proxy, responses=implicit_poly_proxy)
        def create(self, request, *args, **kwargs):
            return Response({})  # pragma: no cover

        @extend_schema(
            request=implicit_poly_proxy,
            responses=implicit_poly_proxy,
            parameters=[OpenApiParameter('id', int, OpenApiParameter.PATH)],
        )
        def partial_update(self, request, *args, **kwargs):
            return Response({})  # pragma: no cover

    explicit_poly_proxy = PolymorphicProxySerializer(
        component_name='MetaPerson',
        serializers={
            'legal': LegalPersonSerializer,
            'natural': NaturalPersonSerializer,
        },
        resource_type_field_name='type',
    )

    class ExplicitPersonViewSet(viewsets.GenericViewSet):
        @extend_schema(request=explicit_poly_proxy, responses=explicit_poly_proxy)
        def create(self, request, *args, **kwargs):
            return Response({})  # pragma: no cover

        @extend_schema(
            request=explicit_poly_proxy,
            responses=explicit_poly_proxy,
            parameters=[OpenApiParameter('id', int, OpenApiParameter.PATH)],
        )
        def partial_update(self, request, *args, **kwargs):
            return Response({})  # pragma: no cover


@pytest.mark.parametrize('viewset', [ImplicitPersonViewSet, ExplicitPersonViewSet])
def test_polymorphic(no_warnings, viewset):
    assert_schema(
        generate_schema('persons', viewset),
        'tests/test_polymorphic.yml'
    )


def test_polymorphic_serializer_as_field_via_extend_schema_field(no_warnings):
    @extend_schema_field(
        PolymorphicProxySerializer(
            component_name='MetaPerson',
            serializers=[LegalPersonSerializer, NaturalPersonSerializer],
            resource_type_field_name='type',
        )
    )
    class XField(serializers.DictField):
        pass  # pragma: no cover

    class XSerializer(serializers.Serializer):
        field = XField()

    @extend_schema(request=XSerializer, responses=XSerializer)
    @api_view(['GET'])
    def view_func(request, format=None):
        pass  # pragma: no cover

    schema = generate_schema('x', view_function=view_func)
    assert 'MetaPerson' in schema['components']['schemas']
    assert 'MetaPerson' in schema['components']['schemas']['X']['properties']['field']['$ref']


def test_polymorphic_serializer_as_method_field_via_extend_schema_field(no_warnings):
    class XSerializer(serializers.Serializer):
        field = serializers.SerializerMethodField()

        @extend_schema_field(
            PolymorphicProxySerializer(
                component_name='MetaPerson',
                serializers=[LegalPersonSerializer, NaturalPersonSerializer],
                resource_type_field_name='type',
            )
        )
        def get_field(self, request):
            pass  # pragma: no cover

    @extend_schema(request=XSerializer, responses=XSerializer)
    @api_view(['GET'])
    def view_func(request, format=None):
        pass  # pragma: no cover

    schema = generate_schema('x', view_function=view_func)
    assert 'MetaPerson' in schema['components']['schemas']
    assert schema['components']['schemas']['X'] == {
        'type': 'object',
        'properties': {
            'field': {'allOf': [{'$ref': '#/components/schemas/MetaPerson'}], 'readOnly': True}
        },
        'required': ['field']
    }


def test_stripped_down_polymorphic_serializer(no_warnings):
    @extend_schema_field(
        PolymorphicProxySerializer(
            component_name='MetaPerson',
            serializers=[LegalPersonSerializer, NaturalPersonSerializer],
            resource_type_field_name=None,
        )
    )
    class XField(serializers.DictField):
        pass  # pragma: no cover

    class XSerializer(serializers.Serializer):
        field = XField()

    @extend_schema(request=XSerializer, responses=XSerializer)
    @api_view(['GET'])
    def view_func(request, format=None):
        pass  # pragma: no cover

    schema = generate_schema('x', view_function=view_func)
    assert schema['components']['schemas']['MetaPerson'] == {'oneOf': [
        {'$ref': '#/components/schemas/LegalPerson'},
        {'$ref': '#/components/schemas/NaturalPerson'}
    ]}


@pytest.mark.parametrize('explicit', [True, False])
def test_many_polymorphic_serializer_extend_schema(no_warnings, explicit):
    if explicit:
        proxy_serializer = serializers.ListSerializer(
            child=PolymorphicProxySerializer(**PROXY_SERIALIZER_PARAMS)
        )
    else:
        proxy_serializer = PolymorphicProxySerializer(**PROXY_SERIALIZER_PARAMS, many=True)

    @extend_schema(request=proxy_serializer, responses=proxy_serializer)
    @api_view(['POST'])
    def view_func(request, format=None):
        pass  # pragma: no cover

    schema = generate_schema('/x/', view_function=view_func)
    assert 'MetaPerson' in schema['components']['schemas']
    op = schema['paths']['/x/']['post']
    assert get_response_schema(op) == {
        'type': 'array',
        'items': {'$ref': '#/components/schemas/MetaPerson'}
    }
    assert get_request_schema(op) == {
        'type': 'array',
        'items': {'$ref': '#/components/schemas/MetaPerson'}
    }


@pytest.mark.parametrize('explicit', [True, False])
def test_many_polymorphic_proxy_serializer_extend_schema_field(no_warnings, explicit):
    if explicit:
        proxy_serializer = serializers.ListField(
            child=PolymorphicProxySerializer(**PROXY_SERIALIZER_PARAMS)
        )
    else:
        proxy_serializer = PolymorphicProxySerializer(**PROXY_SERIALIZER_PARAMS, many=True)

    @extend_schema_field(proxy_serializer)
    class XField(serializers.DictField):
        pass  # pragma: no cover

    class XSerializer(serializers.Serializer):
        field = XField()

    @extend_schema(request=XSerializer, responses=XSerializer)
    @api_view(['POST'])
    def view_func(request, format=None):
        pass  # pragma: no cover

    schema = generate_schema('/x/', view_function=view_func)
    assert 'MetaPerson' in schema['components']['schemas']
    assert schema['components']['schemas']['X'] == {
        'type': 'object',
        'properties': {
            'field': {'type': 'array', 'items': {'$ref': '#/components/schemas/MetaPerson'}}
        },
        'required': ['field']
    }
    op = schema['paths']['/x/']['post']
    assert get_request_schema(op) == {'$ref': '#/components/schemas/X'}
    assert get_response_schema(op) == {'$ref': '#/components/schemas/X'}


def test_polymorphic_proxy_serializer_misusage(no_warnings):
    with pytest.raises(AssertionError):
        PolymorphicProxySerializer(**PROXY_SERIALIZER_PARAMS).data

    with pytest.raises(AssertionError):
        PolymorphicProxySerializer(**PROXY_SERIALIZER_PARAMS).to_representation(None)

    with pytest.raises(AssertionError):
        PolymorphicProxySerializer(**PROXY_SERIALIZER_PARAMS).to_internal_value(None)


@mock.patch('drf_spectacular.settings.spectacular_settings.COMPONENT_SPLIT_REQUEST', True)
@pytest.mark.parametrize('explicit', [True, False])
def test_polymorphic_split_request_with_ro_serializer(no_warnings, explicit):
    class BasicPersonSerializer(serializers.ModelSerializer):
        type = serializers.SerializerMethodField()

        class Meta:
            model = NaturalPerson2
            fields = ('id', 'type')

        def get_type(self, obj) -> str:
            return 'basic'

    if explicit:
        poly_proxy = PolymorphicProxySerializer(
            component_name='MetaPerson',
            serializers={'natural': NaturalPersonSerializer, 'basic': BasicPersonSerializer},
            resource_type_field_name='type',
        )
    else:
        poly_proxy = PolymorphicProxySerializer(
            component_name='MetaPerson',
            serializers=[NaturalPersonSerializer, BasicPersonSerializer],
            resource_type_field_name='type',
        )

    @extend_schema(request=poly_proxy, responses=poly_proxy)
    @api_view(['POST'])
    def view_func(request, format=None):
        pass  # pragma: no cover

    schema = generate_schema('x', view_function=view_func)
    components = schema['components']['schemas']
    assert 'BasicPersonRequest' not in components
    assert components['MetaPerson']['oneOf'] == [
        {'$ref': '#/components/schemas/NaturalPerson'},
        {'$ref': '#/components/schemas/BasicPerson'}
    ]
    assert components['MetaPerson']['discriminator']['mapping'] == {
        'natural': '#/components/schemas/NaturalPerson',
        'basic': '#/components/schemas/BasicPerson'
    }
    assert components['MetaPersonRequest']['oneOf'] == [
        {'$ref': '#/components/schemas/NaturalPersonRequest'},
    ]
    assert components['MetaPersonRequest']['discriminator']['mapping'] == {
        'natural': '#/components/schemas/NaturalPersonRequest',
    }


def test_polymorphic_forced_many_false(no_warnings):
    class XViewSet(viewsets.GenericViewSet):
        @extend_schema(
            responses=PolymorphicProxySerializer(
                component_name='MetaPerson',
                serializers=[NaturalPersonSerializer, LegalPersonSerializer],
                resource_type_field_name='type',
                many=False
            )
        )
        def list(self, request, *args, **kwargs):
            return Response({})  # pragma: no cover

    schema = generate_schema('x', XViewSet)
    assert get_response_schema(schema['paths']['/x/']['get']) == {
        '$ref': '#/components/schemas/MetaPerson'
    }


def test_polymorphic_manual_many(no_warnings):
    mixed_poly = PolymorphicProxySerializer(
        component_name='MetaLegalPerson',
        serializers=[NaturalPersonSerializer, NaturalPersonSerializer(many=True)],
        resource_type_field_name=None,
        many=False,
    )

    class XViewSet(viewsets.GenericViewSet):
        @extend_schema(request=mixed_poly, responses=mixed_poly)
        def create(self, request, *args, **kwargs):
            return Response({})  # pragma: no cover

        @extend_schema(responses=mixed_poly)
        def list(self, request, *args, **kwargs):
            return Response({})  # pragma: no cover

    schema = generate_schema('x', XViewSet)
    response_schema = get_response_schema(schema['paths']['/x/']['post'])
    request_schema = get_request_schema(schema['paths']['/x/']['post'])
    assert response_schema == request_schema == {'$ref': '#/components/schemas/MetaLegalPerson'}
    assert schema['components']['schemas']['MetaLegalPerson'] == {
        'oneOf': [
            {'$ref': '#/components/schemas/NaturalPerson'},
            {'type': 'array', 'items': {'$ref': '#/components/schemas/NaturalPerson'}}
        ]
    }



from drf_spectacular.extensions import OpenApiSerializerFieldExtension
from drf_spectacular.plumbing import build_array_type, is_list_serializer


class RecursiveFieldExtension(OpenApiSerializerFieldExtension):
    target_class = "rest_framework_recursive.fields.RecursiveField"

    def map_serializer_field(self, auto_schema, direction):
        proxied = self.target.proxied

        if is_list_serializer(proxied):
            component = auto_schema.resolve_serializer(proxied.child, direction)
            return build_array_type(component.ref)

        component = auto_schema.resolve_serializer(proxied, direction)
        return component.ref



import sys
import typing

import pytest
from django.urls import path
from rest_framework.decorators import api_view

from drf_spectacular.utils import extend_schema, extend_schema_serializer
from tests import assert_schema, generate_schema


@pytest.mark.contrib('rest_framework_dataclasses')
@pytest.mark.skipif(sys.version_info < (3, 7), reason='dataclass required by package')
def test_rest_framework_dataclasses(no_warnings):
    from dataclasses import dataclass

    from rest_framework_dataclasses.serializers import DataclassSerializer

    @dataclass
    class PersonDetail:
        name: str
        length: int

    @dataclass
    class Person:
        name: str
        length: int
        detail: PersonDetail

    @dataclass
    class Group:
        name: str
        leader: Person
        members: typing.List[Person]

    class GroupSerializer(DataclassSerializer):
        class Meta:
            dataclass = Group

    class GroupSerializer2(DataclassSerializer):
        class Meta:
            dataclass = Group
            ref_name = "CustomGroupNameFromRefName"

    @extend_schema_serializer(component_name='CustomGroupNameFromDecoration')
    @dataclass
    class Group2:
        name: str
        leader: Person
        members: typing.List[Person]

    @extend_schema(responses=GroupSerializer)
    @api_view(['GET'])
    def named(request):
        pass  # pragma: no cover

    @extend_schema(responses=DataclassSerializer(dataclass=Person))
    @api_view(['GET'])
    def anonymous(request):
        pass  # pragma: no cover

    @extend_schema(responses=GroupSerializer2(many=True))
    @api_view(['GET'])
    def custom_name_via_ref(request):
        pass  # pragma: no cover

    @extend_schema(responses=DataclassSerializer(dataclass=Group2))
    @api_view(['GET'])
    def custom_name_via_decoration(request):
        pass  # pragma: no cover

    urlpatterns = [
        path('named', named),
        path('anonymous', anonymous),
        path('custom_name_via_ref', custom_name_via_ref),
        path('custom_name_via_decoration', custom_name_via_decoration)
    ]
    assert_schema(
        generate_schema(None, patterns=urlpatterns),
        'tests/contrib/test_rest_framework_dataclasses.yml'
    )



import contextlib
import functools
import inspect
import sys
from collections import defaultdict
from typing import DefaultDict, List, Optional, Tuple


class GeneratorStats:
    _warn_cache: DefaultDict[str, int] = defaultdict(int)
    _error_cache: DefaultDict[str, int] = defaultdict(int)
    _traces: List[Tuple[Optional[str], Optional[str], str]] = []
    _trace_lineno = False

    _blue = ''
    _red = ''
    _yellow = ''
    _clear = ''

    def __getattr__(self, name):
        if 'silent' not in self.__dict__:
            from drf_spectacular.settings import spectacular_settings
            self.silent = spectacular_settings.DISABLE_ERRORS_AND_WARNINGS
        try:
            return self.__dict__[name]
        except KeyError:
            raise AttributeError(name)

    def __bool__(self):
        return bool(self._warn_cache or self._error_cache)

    @contextlib.contextmanager
    def silence(self):
        self.silent, tmp = True, self.silent
        try:
            yield
        finally:
            self.silent = tmp

    def reset(self):
        self._warn_cache.clear()
        self._error_cache.clear()

    def enable_color(self):
        self._blue = '\033[0;34m'
        self._red = '\033[0;31m'
        self._yellow = '\033[0;33m'
        self._clear = '\033[0m'

    def enable_trace_lineno(self):
        self._trace_lineno = True

    def _get_current_trace(self):
        source_locations = [t for t in self._traces if t[0]]
        if source_locations:
            sourcefile, lineno, _ = source_locations[-1]
            source_location = f'{sourcefile}:{lineno}' if lineno else sourcefile
        else:
            source_location = ''
        breadcrumbs = ' > '.join(t[2] for t in self._traces)
        return source_location, breadcrumbs

    def emit(self, msg, severity):
        assert severity in ['warning', 'error']
        cache = self._warn_cache if severity == 'warning' else self._error_cache

        source_location, breadcrumbs = self._get_current_trace()
        prefix = f'{self._blue}{source_location}: ' if source_location else ''
        prefix += self._yellow if severity == 'warning' else self._red
        prefix += f'{severity.capitalize()}'
        prefix += f' [{breadcrumbs}]: ' if breadcrumbs else ': '

        msg = prefix + self._clear + str(msg)
        if not self.silent and msg not in cache:
            print(msg, file=sys.stderr)
        cache[msg] += 1

    def emit_summary(self):
        if not self.silent and (self._warn_cache or self._error_cache):
            print(
                f'\nSchema generation summary:\n'
                f'Warnings: {sum(self._warn_cache.values())} ({len(self._warn_cache)} unique)\n'
                f'Errors:   {sum(self._error_cache.values())} ({len(self._error_cache)} unique)\n',
                file=sys.stderr
            )


GENERATOR_STATS = GeneratorStats()


def warn(msg, delayed=None):
    if delayed:
        warnings = get_override(delayed, 'warnings', [])
        warnings.append(msg)
        set_override(delayed, 'warnings', warnings)
    else:
        GENERATOR_STATS.emit(msg, 'warning')


def error(msg, delayed=None):
    if delayed:
        errors = get_override(delayed, 'errors', [])
        errors.append(msg)
        set_override(delayed, 'errors', errors)
    else:
        GENERATOR_STATS.emit(msg, 'error')


def reset_generator_stats():
    GENERATOR_STATS.reset()


@contextlib.contextmanager
def add_trace_message(obj):
    """
    Adds a message to be used as a prefix when emitting warnings and errors.
    """
    sourcefile, lineno = _get_source_location(obj)
    GENERATOR_STATS._traces.append((sourcefile, lineno, obj.__name__))
    yield
    GENERATOR_STATS._traces.pop()


@functools.lru_cache(maxsize=1000)
def _get_source_location(obj):
    try:
        sourcefile = inspect.getsourcefile(obj)
    except:  # noqa: E722
        sourcefile = None
    try:
        # This is a rather expensive operation. Only do it when explicitly enabled (CLI)
        # and cache results to speed up some recurring objects like serializers.
        lineno = inspect.getsourcelines(obj)[1] if GENERATOR_STATS._trace_lineno else None
    except:  # noqa: E722
        lineno = None
    return sourcefile, lineno


def has_override(obj, prop):
    if isinstance(obj, functools.partial):
        obj = obj.func
    if not hasattr(obj, '_spectacular_annotation'):
        return False
    if prop not in obj._spectacular_annotation:
        return False
    return True


def get_override(obj, prop, default=None):
    if isinstance(obj, functools.partial):
        obj = obj.func
    if not has_override(obj, prop):
        return default
    return obj._spectacular_annotation[prop]


def set_override(obj, prop, value):
    if not hasattr(obj, '_spectacular_annotation'):
        obj._spectacular_annotation = {}
    elif '_spectacular_annotation' not in obj.__dict__:
        obj._spectacular_annotation = obj._spectacular_annotation.copy()
    obj._spectacular_annotation[prop] = value
    return obj


def get_view_method_names(view, schema=None):
    schema = schema or view.schema
    return [
        item for item in dir(view) if callable(getattr(view, item)) and (
            item in view.http_method_names
            or item in schema.method_mapping.values()
            or item == 'list'
            or hasattr(getattr(view, item), 'mapping')
        )
    ]


def isolate_view_method(view, method_name):
    """
    Prevent modifying a view method which is derived from other views. Changes to
    a derived method would leak into the view where the method originated from.
    Break derivation by wrapping the method and explicitly setting it on the view.
    """
    method = getattr(view, method_name)
    # no isolation is required if the view method is not derived.
    # @api_view is a special case that also breaks isolation. It proxies all view
    # methods through a single handler function, which then also requires isolation.
    if method_name in view.__dict__ and method.__name__ != 'handler':
        return method

    @functools.wraps(method)
    def wrapped_method(self, request, *args, **kwargs):
        return method(self, request, *args, **kwargs)

    # wraps() will only create a shallow copy of method.__dict__. Updates to "kwargs"
    # via @extend_schema would leak to the original method. Isolate by creating a copy.
    if hasattr(method, 'kwargs'):
        wrapped_method.kwargs = method.kwargs.copy()

    setattr(view, method_name, wrapped_method)
    return wrapped_method


def cache(user_function):
    """ simple polyfill for python < 3.9 """
    return functools.lru_cache(maxsize=None)(user_function)



import pytest
from django.db import models
from rest_framework import mixins, routers, serializers, viewsets
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.test import APIClient

from drf_spectacular.generators import SchemaGenerator
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import extend_schema, extend_schema_view
from tests import assert_schema


class ESVModel(models.Model):
    pass


class ESVSerializer(serializers.ModelSerializer):
    class Meta:
        model = ESVModel
        fields = '__all__'


class DualMethodActionParamsSerializer(serializers.Serializer):
    message = serializers.CharField()


@extend_schema(tags=['global-tag'])
@extend_schema_view(
    list=extend_schema(description='view list description'),
    retrieve=extend_schema(description='view retrieve description'),
    extended_action=extend_schema(description='view extended action description'),
    raw_action=extend_schema(description='view raw action description'),
    dual_method_action=[
        extend_schema(parameters=[DualMethodActionParamsSerializer], methods=['GET']),
        extend_schema(request=DualMethodActionParamsSerializer, methods=['POST']),
    ]
)
class XViewSet(mixins.ListModelMixin, mixins.RetrieveModelMixin, viewsets.GenericViewSet):
    queryset = ESVModel.objects.all()
    serializer_class = ESVSerializer

    @extend_schema(tags=['custom-retrieve-tag'])
    def retrieve(self, request, *args, **kwargs):
        return super().retrieve(request, *args, **kwargs)

    @extend_schema(responses=OpenApiTypes.DATE)
    @action(detail=False)
    def extended_action(self, request):
        return Response('2020-10-31')

    @action(detail=False, methods=['GET'])
    def raw_action(self, request):
        return Response('2019-03-01')

    @extend_schema(description='view dual method action description')
    @action(detail=False, methods=['GET', 'POST'])
    def dual_method_action(self, request):
        if request.method == 'POST':
            data = request.data
        else:
            data = request.query_params
        return Response(data['message'])


# view to make sure there is no cross-talk
class YViewSet(viewsets.ModelViewSet):
    serializer_class = ESVSerializer
    queryset = ESVModel.objects.all()


# view to make sure that schema applied to a subclass does not affect its parent.
@extend_schema_view(
    list=extend_schema(exclude=True),
    retrieve=extend_schema(description='overridden description for child only'),
    extended_action=extend_schema(responses={200: {'type': 'string', 'pattern': r'^[0-9]{4}(?:-[0-9]{2}){2}$'}}),
    raw_action=extend_schema(summary="view raw action summary"),
)
class ZViewSet(XViewSet):
    @extend_schema(tags=['child-tag'])
    @action(detail=False, methods=['GET'])
    def raw_action(self, request):
        return Response('2019-03-01')  # pragma: no cover


router = routers.SimpleRouter()
router.register('x', XViewSet)
router.register('y', YViewSet)
router.register('z', ZViewSet)
urlpatterns = router.urls


@pytest.mark.urls(__name__)
def test_extend_schema_view(no_warnings):
    assert_schema(
        SchemaGenerator().get_schema(request=None, public=True),
        'tests/test_extend_schema_view.yml'
    )


@pytest.mark.urls(__name__)
@pytest.mark.django_db
def test_extend_schema_view_call_transparency(no_warnings):
    ESVModel.objects.create()

    response = APIClient().get('/x/')
    assert response.status_code == 200
    assert response.content == b'[{"id":1}]'
    response = APIClient().get('/x/1/')
    assert response.status_code == 200
    assert response.content == b'{"id":1}'
    response = APIClient().get('/x/extended_action/')
    assert response.status_code == 200
    assert response.content == b'"2020-10-31"'
    response = APIClient().get('/x/raw_action/')
    assert response.status_code == 200
    assert response.content == b'"2019-03-01"'
    response = APIClient().get('/x/dual_method_action/', {'message': 'foo bar'})
    assert response.status_code == 200
    assert response.content == b'"foo bar"'
    response = APIClient().post('/x/dual_method_action/', {'message': 'foo bar'})
    assert response.status_code == 200
    assert response.content == b'"foo bar"'



import typing
from enum import Enum
from unittest import mock

import pytest
from django import __version__ as DJANGO_VERSION
from rest_framework import generics, mixins, serializers, viewsets
from rest_framework.decorators import action
from rest_framework.views import APIView

try:
    from django.db.models.enums import TextChoices
except ImportError:
    TextChoices = object  # type: ignore  # django < 3.0 handling

from drf_spectacular.plumbing import list_hash, load_enum_name_overrides
from drf_spectacular.utils import OpenApiParameter, extend_schema
from tests import assert_schema, generate_schema

language_choices = (
    ('en', 'en'),
    ('es', 'es'),
    ('ru', 'ru'),
    ('cn', 'cn'),
)

blank_null_language_choices = (
    ('en', 'en'),
    ('es', 'es'),
    ('ru', 'ru'),
    ('cn', 'cn'),
    ('', 'not provided'),
    (None, 'unknown'),
)

vote_choices = (
    (1, 'Positive'),
    (0, 'Neutral'),
    (-1, 'Negative'),
)

language_list = ['en']


class LanguageEnum(Enum):
    EN = 'en'


class LanguageStrEnum(str, Enum):
    EN = 'en'


class LanguageChoices(TextChoices):
    EN = 'en'


blank_null_language_list = ['en', '', None]


class BlankNullLanguageEnum(Enum):
    EN = 'en'
    BLANK = ''
    NULL = None


class BlankNullLanguageStrEnum(str, Enum):
    EN = 'en'
    BLANK = ''
    # These will still be included since the values get cast to strings so 'None' != None
    NULL = None


class BlankNullLanguageChoices(TextChoices):
    EN = 'en'
    BLANK = ''
    # These will still be included since the values get cast to strings so 'None' != None
    NULL = None


class ASerializer(serializers.Serializer):
    language = serializers.ChoiceField(choices=language_choices)
    vote = serializers.ChoiceField(choices=vote_choices)


class BSerializer(serializers.Serializer):
    language = serializers.ChoiceField(choices=language_choices, allow_blank=True, allow_null=True)


class AViewset(mixins.ListModelMixin, viewsets.GenericViewSet):
    serializer_class = ASerializer

    @extend_schema(responses=BSerializer)
    @action(detail=False, serializer_class=BSerializer)
    def selection(self, request):
        pass  # pragma: no cover


def test_postprocessing(no_warnings):
    schema = generate_schema('a', AViewset)
    assert_schema(schema, 'tests/test_postprocessing.yml')


@mock.patch(
    'drf_spectacular.settings.spectacular_settings.ENUM_ADD_EXPLICIT_BLANK_NULL_CHOICE', False
)
def test_no_blank_and_null_in_enum_choices(no_warnings):
    schema = generate_schema('a', AViewset)
    assert 'NullEnum' not in schema['components']['schemas']
    assert 'BlankEnum' not in schema['components']['schemas']
    assert 'oneOf' not in schema['components']['schemas']['B']['properties']['language']


@mock.patch('drf_spectacular.settings.spectacular_settings.ENUM_NAME_OVERRIDES', {
    'LanguageEnum': 'tests.test_postprocessing.language_choices'
})
def test_global_enum_naming_override(no_warnings, clear_caches):
    # the override will prevent the warning for multiple names
    class XSerializer(serializers.Serializer):
        foo = serializers.ChoiceField(choices=language_choices)
        bar = serializers.ChoiceField(choices=language_choices)

    class XView(generics.RetrieveAPIView):
        serializer_class = XSerializer

    schema = generate_schema('/x', view=XView)
    assert 'LanguageEnum' in schema['components']['schemas']['X']['properties']['foo']['$ref']
    assert 'LanguageEnum' in schema['components']['schemas']['X']['properties']['bar']['$ref']
    assert len(schema['components']['schemas']) == 2


@mock.patch('drf_spectacular.settings.spectacular_settings.ENUM_NAME_OVERRIDES', {
    'LanguageEnum': 'tests.test_postprocessing.blank_null_language_choices'
})
def test_global_enum_naming_override_with_blank_and_none(no_warnings, clear_caches):
    """Test that choices with blank values can still have their name overridden."""
    class XSerializer(serializers.Serializer):
        foo = serializers.ChoiceField(choices=blank_null_language_choices)
        bar = serializers.ChoiceField(choices=blank_null_language_choices)

    class XView(generics.RetrieveAPIView):
        serializer_class = XSerializer

    schema = generate_schema('/x', view=XView)
    foo_data = schema['components']['schemas']['X']['properties']['foo']
    bar_data = schema['components']['schemas']['X']['properties']['bar']

    assert len(foo_data['oneOf']) == 3
    assert len(bar_data['oneOf']) == 3

    foo_ref_values = [ref_object['$ref'] for ref_object in foo_data['oneOf']]
    bar_ref_values = [ref_object['$ref'] for ref_object in bar_data['oneOf']]

    assert foo_ref_values == [
        '#/components/schemas/LanguageEnum',
        '#/components/schemas/BlankEnum',
        '#/components/schemas/NullEnum'
    ]
    assert bar_ref_values == [
        '#/components/schemas/LanguageEnum',
        '#/components/schemas/BlankEnum',
        '#/components/schemas/NullEnum'
    ]


def test_enum_name_reuse_warning(capsys):
    class XSerializer(serializers.Serializer):
        foo = serializers.ChoiceField(choices=language_choices)
        bar = serializers.ChoiceField(choices=language_choices)

    class XView(generics.RetrieveAPIView):
        serializer_class = XSerializer

    generate_schema('/x', view=XView)
    assert 'encountered multiple names for the same choice set' in capsys.readouterr().err


def test_enum_collision_without_override(capsys):
    class XSerializer(serializers.Serializer):
        foo = serializers.ChoiceField(choices=[('A', 'A')])

    class YSerializer(serializers.Serializer):
        foo = serializers.ChoiceField(choices=[('A', 'A'), ('B', 'B')])

    class ZSerializer(serializers.Serializer):
        foo = serializers.ChoiceField(choices=[('A', 'A'), ('B', 'B')])

    class XAPIView(APIView):
        @extend_schema(responses=ZSerializer)
        def get(self, request):
            pass  # pragma: no cover

        @extend_schema(request=XSerializer, responses=YSerializer)
        def post(self, request):
            pass  # pragma: no cover

    generate_schema('x', view=XAPIView)
    assert 'enum naming encountered a non-optimally resolvable' in capsys.readouterr().err


def test_resolvable_enum_collision(no_warnings):
    class XSerializer(serializers.Serializer):
        foo = serializers.ChoiceField(choices=[('A', 'A')])

    class YSerializer(serializers.Serializer):
        foo = serializers.ChoiceField(choices=[('A', 'A'), ('B', 'B')])

    class XAPIView(APIView):
        @extend_schema(request=XSerializer, responses=YSerializer)
        def post(self, request):
            pass  # pragma: no cover

    schema = generate_schema('x', view=XAPIView)
    assert 'XFooEnum' in schema['components']['schemas']
    assert 'YFooEnum' in schema['components']['schemas']


@mock.patch('drf_spectacular.settings.spectacular_settings.COMPONENT_SPLIT_PATCH', True)
@mock.patch('drf_spectacular.settings.spectacular_settings.COMPONENT_SPLIT_REQUEST', True)
def test_enum_resolvable_collision_with_patched_and_request_splits():
    class XSerializer(serializers.Serializer):
        foo = serializers.ChoiceField(choices=[('A', 'A')])

    class YSerializer(serializers.Serializer):
        foo = serializers.ChoiceField(choices=[('A', 'A'), ('B', 'B')])

    class XViewset(viewsets.GenericViewSet):
        @extend_schema(request=XSerializer, responses=YSerializer)
        def create(self, request):
            pass  # pragma: no cover

        @extend_schema(
            request=XSerializer,
            responses=YSerializer,
            parameters=[OpenApiParameter('id', int, OpenApiParameter.PATH)]
        )
        def partial_update(self, request):
            pass  # pragma: no cover

    schema = generate_schema('/x', XViewset)
    components = schema['components']['schemas']
    assert 'XFooEnum' in components and 'YFooEnum' in components
    assert '/XFooEnum' in components['XRequest']['properties']['foo']['$ref']
    assert '/XFooEnum' in components['PatchedXRequest']['properties']['foo']['$ref']


def test_enum_override_variations(no_warnings):
    enum_override_variations = ['language_list', 'LanguageEnum', 'LanguageStrEnum']
    if DJANGO_VERSION > '3':
        enum_override_variations += ['LanguageChoices', 'LanguageChoices.choices']

    for variation in enum_override_variations:
        with mock.patch(
            'drf_spectacular.settings.spectacular_settings.ENUM_NAME_OVERRIDES',
            {'LanguageEnum': f'tests.test_postprocessing.{variation}'}
        ):
            load_enum_name_overrides.cache_clear()
            assert list_hash(['en']) in load_enum_name_overrides()


def test_enum_override_variations_with_blank_and_null(no_warnings):
    enum_override_variations = [
        'blank_null_language_list',
        'BlankNullLanguageEnum',
        ('BlankNullLanguageStrEnum', ['en', 'None'])
    ]
    if DJANGO_VERSION > '3':
        enum_override_variations += [
            ('BlankNullLanguageChoices', ['en', 'None']),
            ('BlankNullLanguageChoices.choices', ['en', 'None'])
        ]

    for variation in enum_override_variations:
        expected_hashed_keys = ['en']
        if isinstance(variation, (list, tuple, )):
            variation, expected_hashed_keys = variation
        with mock.patch(
            'drf_spectacular.settings.spectacular_settings.ENUM_NAME_OVERRIDES',
            {'LanguageEnum': f'tests.test_postprocessing.{variation}'}
        ):
            load_enum_name_overrides.cache_clear()
            # Should match after None and blank strings are removed
            assert list_hash(expected_hashed_keys) in load_enum_name_overrides()


@mock.patch('drf_spectacular.settings.spectacular_settings.ENUM_NAME_OVERRIDES', {
    'LanguageEnum': 'tests.test_postprocessing.NOTEXISTING'
})
def test_enum_override_loading_fail(capsys, clear_caches):
    load_enum_name_overrides()
    assert 'unable to load choice override for LanguageEnum' in capsys.readouterr().err


@pytest.mark.skipif(DJANGO_VERSION < '3', reason='Not available before Django 3.0')
def test_textchoice_annotation(no_warnings):
    class QualityChoices(TextChoices):
        GOOD = 'GOOD'
        BAD = 'BAD'

    class XSerializer(serializers.Serializer):
        quality_levels = serializers.SerializerMethodField()

        def get_quality_levels(self, obj) -> typing.List[QualityChoices]:
            return [QualityChoices.GOOD, QualityChoices.BAD]  # pragma: no cover

    class XAPIView(APIView):
        @extend_schema(responses=XSerializer)
        def get(self, request):
            pass  # pragma: no cover

    schema = generate_schema('x', view=XAPIView)
    assert 'QualityLevelsEnum' in schema['components']['schemas']
    assert schema['components']['schemas']['X']['properties']['quality_levels'] == {
        'type': 'array',
        'items': {'$ref': '#/components/schemas/QualityLevelsEnum'},
        'readOnly': True
    }



import uuid
from typing import Optional

from django.db import models
from rest_framework import serializers, viewsets
from rest_framework.authentication import TokenAuthentication
from rest_framework.decorators import action
from rest_framework.permissions import IsAuthenticatedOrReadOnly
from rest_framework.response import Response

from tests import assert_schema, generate_schema


class Album(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    title = models.CharField(max_length=100)
    genre = models.CharField(
        choices=(('POP', 'Pop'), ('ROCK', 'Rock')),
        max_length=10
    )
    year = models.IntegerField()
    released = models.BooleanField()


class Song(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    album = models.ForeignKey(Album, on_delete=models.CASCADE)

    title = models.CharField(max_length=100)
    length = models.IntegerField()


class SongSerializer(serializers.ModelSerializer):
    top10 = serializers.SerializerMethodField()

    class Meta:
        fields = ['id', 'title', 'length', 'top10']
        model = Song

    def get_top10(self) -> Optional[bool]:
        return True  # pragma: no cover


class AlbumSerializer(serializers.ModelSerializer):
    songs = SongSerializer(many=True, read_only=True)
    single = SongSerializer(read_only=True)

    class Meta:
        fields = '__all__'
        model = Album


class LikeSerializer(serializers.Serializer):
    def save(self, *args, **kwargs):
        pass  # pragma: no cover


class AlbumModelViewset(viewsets.ModelViewSet):
    authentication_classes = [TokenAuthentication]
    permission_classes = [IsAuthenticatedOrReadOnly]
    serializer_class = AlbumSerializer
    queryset = Album.objects.none()

    @action(detail=True, methods=['POST'], serializer_class=LikeSerializer)
    def like(self, request):
        return Response(self.get_serializer().data)  # pragma: no cover

    def create(self, request, *args, **kwargs):
        """
        Special documentation about creating albums

        There is even more info here
        """
        return super().create(request, *args, **kwargs)  # pragma: no cover


def test_basic(no_warnings):
    assert_schema(
        generate_schema('albums', AlbumModelViewset),
        'tests/test_basic.yml'
    )



import re
from importlib import reload
from unittest import mock

import pytest
from django.urls import include, path
from rest_framework import viewsets

from tests import assert_schema, generate_schema
from tests.models import SimpleModel, SimpleSerializer

transforms = [
    # User model first_name differences
    lambda x: re.sub(r'(first_name:\n *type: string\n *maxLength:) 30', r'\g<1> 150', x, re.M),
]


@pytest.mark.contrib('dj_rest_auth', 'allauth')
@mock.patch('drf_spectacular.settings.spectacular_settings.SCHEMA_PATH_PREFIX', '')
def test_rest_auth(no_warnings):
    urlpatterns = [
        path('rest-auth/', include('dj_rest_auth.urls')),
        path('rest-auth/registration/', include('dj_rest_auth.registration.urls')),
    ]
    schema = generate_schema(None, patterns=urlpatterns)
    assert_schema(
        schema, 'tests/contrib/test_rest_auth.yml', transforms=transforms
    )


@pytest.mark.contrib('dj_rest_auth', 'allauth', 'rest_framework_simplejwt')
@mock.patch('drf_spectacular.settings.spectacular_settings.SCHEMA_PATH_PREFIX', '')
def test_rest_auth_token(no_warnings, settings):
    settings.REST_USE_JWT = True
    # flush module import cache to re-evaluate conditional import
    import dj_rest_auth.urls
    reload(dj_rest_auth.urls)

    urlpatterns = [
        # path('rest-auth/', include(urlpatterns)),
        path('rest-auth/', include('dj_rest_auth.urls')),
        path('rest-auth/registration/', include('dj_rest_auth.registration.urls')),
    ]

    schema = generate_schema(None, patterns=urlpatterns)

    assert_schema(
        schema, 'tests/contrib/test_rest_auth_token.yml', transforms=transforms
    )


@pytest.mark.contrib('dj_rest_auth', 'rest_framework_simplejwt')
@mock.patch('django.conf.settings.JWT_AUTH_COOKIE', 'jwt-session', create=True)
def test_rest_auth_simplejwt_cookie(no_warnings):
    from dj_rest_auth.jwt_auth import JWTCookieAuthentication

    class XViewset(viewsets.ModelViewSet):
        serializer_class = SimpleSerializer
        queryset = SimpleModel.objects.all()
        authentication_classes = [JWTCookieAuthentication]

    schema = generate_schema('/x', XViewset)
    assert schema['paths']['/x/']['get']['security'] == [
        {'jwtHeaderAuth': []}, {'jwtCookieAuth': []}, {}
    ]
    assert schema['components']['securitySchemes'] == {
        'jwtCookieAuth': {'type': 'apiKey', 'in': 'cookie', 'name': 'jwt-session'},
        'jwtHeaderAuth': {'type': 'http', 'scheme': 'bearer', 'bearerFormat': 'JWT'}
    }



from rest_framework.utils.model_meta import get_field_info

from drf_spectacular.drainage import warn
from drf_spectacular.extensions import OpenApiSerializerExtension, OpenApiSerializerFieldExtension
from drf_spectacular.plumbing import (
    ResolvedComponent, build_array_type, build_object_type, follow_field_source, get_doc,
)


def build_point_schema():
    return {
        "type": "array",
        "items": {"type": "number", "format": "float"},
        "example": [12.9721, 77.5933],
        "minItems": 2,
        "maxItems": 3,
    }


def build_linestring_schema():
    return {
        "type": "array",
        "items": build_point_schema(),
        "example": [[22.4707, 70.0577], [12.9721, 77.5933]],
        "minItems": 2,
    }


def build_polygon_schema():
    return {
        "type": "array",
        "items": {**build_linestring_schema(), "minItems": 4},
        "example": [
            [
                [0.0, 0.0],
                [0.0, 50.0],
                [50.0, 50.0],
                [50.0, 0.0],
                [0.0, 0.0],
            ],
        ]
    }


def build_geo_container_schema(name, coords):
    return build_object_type(
        properties={
            "type": {"type": "string", "enum": [name]},
            "coordinates": coords,
        }
    )


def build_point_geo_schema():
    return build_geo_container_schema("Point", build_point_schema())


def build_linestring_geo_schema():
    return build_geo_container_schema("LineString", build_linestring_schema())


def build_polygon_geo_schema():
    return build_geo_container_schema("Polygon", build_polygon_schema())


def build_geometry_geo_schema():
    return {
        'oneOf': [
            build_point_geo_schema(),
            build_linestring_geo_schema(),
            build_polygon_geo_schema(),
        ]
    }


def build_bbox_schema():
    return {
        "type": "array",
        "items": {"type": "number"},
        "minItems": 4,
        "maxItems": 4,
        "example": [12.9721, 77.5933, 12.9721, 77.5933],
    }


def build_geo_schema(model_field):
    from django.contrib.gis.db import models

    if isinstance(model_field, models.PointField):
        return build_point_geo_schema()
    elif isinstance(model_field, models.LineStringField):
        return build_linestring_geo_schema()
    elif isinstance(model_field, models.PolygonField):
        return build_polygon_geo_schema()
    elif isinstance(model_field, models.MultiPointField):
        return build_geo_container_schema(
            "MultiPoint", build_array_type(build_point_schema())
        )
    elif isinstance(model_field, models.MultiLineStringField):
        return build_geo_container_schema(
            "MultiLineString", build_array_type(build_linestring_schema())
        )
    elif isinstance(model_field, models.MultiPolygonField):
        return build_geo_container_schema(
            "MultiPolygon", build_array_type(build_polygon_schema())
        )
    elif isinstance(model_field, models.GeometryCollectionField):
        return build_geo_container_schema(
            "GeometryCollection", build_array_type(build_geometry_geo_schema())
        )
    elif isinstance(model_field, models.GeometryField):
        return build_geometry_geo_schema()
    else:
        warn("Encountered unknown GIS geometry field")
        return {}


def map_geo_field(serializer, geo_field_name):
    from rest_framework_gis.fields import GeometrySerializerMethodField

    field = serializer.fields[geo_field_name]
    if isinstance(field, GeometrySerializerMethodField):
        warn("Geometry generation for GeometrySerializerMethodField is not supported.")
        return {}
    model_field = get_field_info(serializer.Meta.model).fields[geo_field_name]
    return build_geo_schema(model_field)


def _inject_enum_collision_fix(collection):
    from drf_spectacular.settings import spectacular_settings
    if not collection and 'GisFeatureEnum' not in spectacular_settings.ENUM_NAME_OVERRIDES:
        spectacular_settings.ENUM_NAME_OVERRIDES['GisFeatureEnum'] = ('Feature',)
    if collection and 'GisFeatureCollectionEnum' not in spectacular_settings.ENUM_NAME_OVERRIDES:
        spectacular_settings.ENUM_NAME_OVERRIDES['GisFeatureCollectionEnum'] = ('FeatureCollection',)


class GeoFeatureModelSerializerExtension(OpenApiSerializerExtension):
    target_class = 'rest_framework_gis.serializers.GeoFeatureModelSerializer'
    match_subclasses = True

    def map_serializer(self, auto_schema, direction):
        _inject_enum_collision_fix(collection=False)

        base_schema = auto_schema._map_serializer(self.target, direction, bypass_extensions=True)
        return self.map_geo_feature_model_serializer(self.target, base_schema)

    def map_geo_feature_model_serializer(self, serializer, base_schema):
        from rest_framework_gis.serializers import GeoFeatureModelSerializer

        geo_properties = {
            "type": {"type": "string", "enum": ["Feature"]}
        }
        if serializer.Meta.id_field:
            geo_properties["id"] = base_schema["properties"].pop(serializer.Meta.id_field)

        geo_properties["geometry"] = map_geo_field(serializer, serializer.Meta.geo_field)
        base_schema["properties"].pop(serializer.Meta.geo_field)

        if serializer.Meta.auto_bbox or serializer.Meta.bbox_geo_field:
            geo_properties["bbox"] = build_bbox_schema()
            base_schema["properties"].pop(serializer.Meta.bbox_geo_field, None)

        # only expose if description comes from the user
        description = base_schema.pop('description', None)
        if description == get_doc(GeoFeatureModelSerializer):
            description = None

        # ignore this aspect for now
        base_schema.pop('required', None)

        # nest remaining fields under property "properties"
        geo_properties["properties"] = base_schema

        return build_object_type(
            properties=geo_properties,
            description=description,
        )


class GeoFeatureModelListSerializerExtension(OpenApiSerializerExtension):
    target_class = 'rest_framework_gis.serializers.GeoFeatureModelListSerializer'

    def map_serializer(self, auto_schema, direction):
        _inject_enum_collision_fix(collection=True)

        # build/retrieve feature component generated by GeoFeatureModelSerializerExtension.
        # wrap the ref in the special list structure and build another component based on that.
        feature_component = auto_schema.resolve_serializer(self.target.child, direction)
        collection_schema = build_object_type(
            properties={
                "type": {"type": "string", "enum": ["FeatureCollection"]},
                "features": build_array_type(feature_component.ref)
            }
        )
        list_component = ResolvedComponent(
            name=f'{feature_component.name}List',
            type=ResolvedComponent.SCHEMA,
            object=self.target.child,
            schema=collection_schema
        )
        auto_schema.registry.register_on_missing(list_component)
        return list_component.ref


class GeometryFieldExtension(OpenApiSerializerFieldExtension):
    target_class = 'rest_framework_gis.fields.GeometryField'
    match_subclasses = True

    def map_serializer_field(self, auto_schema, direction):
        # running this extension for GeoFeatureModelSerializer's geo_field is superfluous
        # as above extension already handles that individually. We run it anyway because
        # robustly checking the proper condition is harder.
        try:
            model = self.target.parent.Meta.model
            model_field = follow_field_source(model, self.target.source.split('.'))
            return build_geo_schema(model_field)
        except:  # noqa: E722
            warn(f'Encountered an issue resolving field {self.target}. defaulting to generic object.')
            return {}



import inspect

import pytest

from drf_spectacular.drainage import GENERATOR_STATS


def test_known_attribute_access_succeeds():
    assert hasattr(GENERATOR_STATS, 'silent')


def test_unknown_attribute_access_fails():
    with pytest.raises(AttributeError):
        getattr(GENERATOR_STATS, '__spam__')


def test_inspect_unwrap():
    assert inspect.unwrap(GENERATOR_STATS) is GENERATOR_STATS



from rest_framework import serializers

from drf_spectacular.extensions import (
    OpenApiSerializerExtension, OpenApiSerializerFieldExtension, OpenApiViewExtension
)
from drf_spectacular.plumbing import build_basic_type
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import OpenApiParameter, extend_schema, extend_schema_field


class Fix1(OpenApiViewExtension):
    target_class = 'oscarapi.views.root.api_root'

    def view_replacement(self):
        return extend_schema(responses=OpenApiTypes.OBJECT)(self.target_class)


class Fix2(OpenApiViewExtension):
    target_class = 'oscarapi.views.product.ProductAvailability'

    def view_replacement(self):
        from oscarapi.serializers.product import AvailabilitySerializer

        class Fixed(self.target_class):
            serializer_class = AvailabilitySerializer
        return Fixed


class Fix3(OpenApiViewExtension):
    target_class = 'oscarapi.views.product.ProductPrice'

    def view_replacement(self):
        from oscarapi.serializers.checkout import PriceSerializer

        class Fixed(self.target_class):
            serializer_class = PriceSerializer
        return Fixed


class Fix4(OpenApiViewExtension):
    target_class = 'oscarapi.views.checkout.UserAddressDetail'

    def view_replacement(self):
        from oscar.apps.address.models import UserAddress

        class Fixed(self.target_class):
            queryset = UserAddress.objects.none()
        return Fixed


class Fix5(OpenApiViewExtension):
    target_class = 'oscarapi.views.product.CategoryList'

    def view_replacement(self):
        class Fixed(self.target_class):
            @extend_schema(parameters=[
                OpenApiParameter(name='breadcrumbs', type=OpenApiTypes.STR, location=OpenApiParameter.PATH)
            ])
            def get(self, request, *args, **kwargs):
                pass

        return Fixed


class Fix6(OpenApiSerializerExtension):
    target_class = 'oscarapi.serializers.checkout.OrderSerializer'

    def map_serializer(self, auto_schema, direction):
        from oscarapi.serializers.checkout import OrderOfferDiscountSerializer, OrderVoucherOfferSerializer

        class Fixed(self.target_class):
            @extend_schema_field(OrderOfferDiscountSerializer(many=True))
            def get_offer_discounts(self):
                pass

            @extend_schema_field(OpenApiTypes.URI)
            def get_payment_url(self):
                pass

            @extend_schema_field(OrderVoucherOfferSerializer(many=True))
            def get_voucher_discounts(self):
                pass

        return auto_schema._map_serializer(Fixed, direction)


class Fix7(OpenApiSerializerFieldExtension):
    target_class = 'oscarapi.serializers.fields.CategoryField'

    def map_serializer_field(self, auto_schema, direction):
        return build_basic_type(OpenApiTypes.STR)


class Fix8(OpenApiSerializerFieldExtension):
    target_class = 'oscarapi.serializers.fields.AttributeValueField'

    def map_serializer_field(self, auto_schema, direction):
        return {
            'oneOf': [
                build_basic_type(OpenApiTypes.STR),
            ]
        }


class Fix9(OpenApiSerializerExtension):
    target_class = 'oscarapi.serializers.basket.BasketSerializer'

    def map_serializer(self, auto_schema, direction):
        class Fixed(self.target_class):
            is_tax_known = serializers.SerializerMethodField()

            def get_is_tax_known(self) -> bool:
                pass

        return auto_schema._map_serializer(Fixed, direction)


class Fix10(Fix9):
    target_class = 'oscarapi.serializers.basket.BasketLineSerializer'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re
import shutil
import sys

from setuptools import find_namespace_packages, setup

name = 'drf-spectacular'
package = 'drf_spectacular'
description = 'Sane and flexible OpenAPI 3 schema generation for Django REST framework'
url = 'https://github.com/tfranzel/drf-spectacular'
author = 'T. Franzel'
author_email = 'tfranzel@gmail.com'
license = 'BSD'

with open('README.rst') as readme:
    long_description = readme.read()

with open('requirements/base.txt') as fh:
    requirements = [r for r in fh.read().split('\n') if not r.startswith('#')]


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("^__version__ = ['\"]([^'\"]+)['\"]",
                     init_py, re.MULTILINE).group(1)


version = get_version(package)


if sys.argv[-1] == 'publish':
    if os.system("pip freeze | grep twine"):
        print("twine not installed.\nUse `pip install twine`.\nExiting.")
        sys.exit(1)
    os.system("python setup.py sdist bdist_wheel")
    if os.system("twine check dist/*"):
        print("twine check failed. Packages might be outdated.")
        print("Try using `pip install -U twine wheel`.\nExiting.")
        sys.exit(1)
    if os.system("twine upload dist/*"):
        print("failed to upload package")
        sys.exit(1)
    if os.environ.get('CI'):
        os.system("git config user.name github-actions")
        os.system("git config user.email github-actions@github.com")
    os.system(f"git tag -a {version} -m 'version {version}'")
    if os.system("git push --tags"):
        print("failed pushing release tag")
        sys.exit(1)
    shutil.rmtree('dist')
    shutil.rmtree('build')
    shutil.rmtree('drf_spectacular.egg-info')
    sys.exit()


setup(
    name=name,
    version=version,
    url=url,
    license=license,
    description=description,
    long_description=long_description,
    long_description_content_type='text/x-rst',
    author=author,
    author_email=author_email,
    packages=[p for p in find_namespace_packages(exclude=('tests*',)) if p.startswith(package)],
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=requirements,
    extras_require={
        "offline": ["drf-spectacular-sidecar"],
        "sidecar": ["drf-spectacular-sidecar"],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.2',
        'Framework :: Django :: 4.0',
        'Framework :: Django :: 4.1',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Documentation',
        'Topic :: Software Development :: Code Generators',
    ],
    project_urls={
        'Source': 'https://github.com/tfranzel/drf-spectacular',
        'Documentation': 'https://drf-spectacular.readthedocs.io',
    },
)



from unittest import mock

from rest_framework import serializers, viewsets
from rest_framework.decorators import action

from drf_spectacular.utils import OpenApiCallback, OpenApiResponse, extend_schema, inline_serializer
from tests import assert_schema, generate_schema
from tests.models import SimpleModel, SimpleSerializer


class EventSerializer(serializers.Serializer):
    id = serializers.CharField(read_only=True)
    change = serializers.CharField()
    external_id = serializers.CharField(write_only=True)


class XViewset(viewsets.ReadOnlyModelViewSet):
    queryset = SimpleModel.objects.none()
    serializer_class = SimpleSerializer

    @extend_schema(
        request=inline_serializer(
            name='SubscribeSerializer',
            fields={'callbackUrl': serializers.URLField()},
        ),
        responses=None,
        callbacks=[
            OpenApiCallback(
                name='SubscriptionEvent',
                path='{$request.body#/callbackUrl}',
                decorator=extend_schema(
                    summary="some summary",
                    description='pushes events to callbackUrl as "application/x-www-form-urlencoded"',
                    request={
                        'application/x-www-form-urlencoded': EventSerializer,
                    },
                    responses={
                        200: OpenApiResponse(description='event was successfully received'),
                        '4XX': OpenApiResponse(description='event will be retried shortly'),
                    },
                ),
            ),
        ]
    )
    @action(detail=False, methods=['POST'])
    def subscription(self):
        pass  # pragma: no cover

    @extend_schema(
        request=inline_serializer(
            name='HealthSerializer',
            fields={'callbackUrl': serializers.URLField()},
        ),
        responses=None,
        callbacks=[
            OpenApiCallback(
                name='HealthEvent',
                path='{$request.body#/callbackUrl}',
                decorator={
                    'post': extend_schema(
                        request=EventSerializer,
                        responses=OpenApiResponse(description='status new ack'),
                    ),
                    'delete': extend_schema(
                        deprecated=True,
                        responses={200: OpenApiResponse(description='status expiration')},
                    ),
                    'put': extend_schema(
                        request=EventSerializer,
                        responses=EventSerializer,
                    ),
                    # raw schema
                    'patch': {
                        'requestBody': {'content': {'application/yaml': {'schema': {'type': 'integer'}}}},
                        'responses': {'200': {'description': 'Raw schema'}}
                    },
                },
            ),
        ]
    )
    @action(detail=False, methods=['POST'])
    def health(self):
        pass  # pragma: no cover


def test_callbacks(no_warnings):
    assert_schema(
        generate_schema('/x', XViewset),
        'tests/test_callbacks.yml'
    )


@mock.patch('drf_spectacular.settings.spectacular_settings.COMPONENT_SPLIT_REQUEST', True)
def test_callbacks_split_request(no_warnings):
    assert_schema(
        generate_schema('/x', XViewset),
        'tests/test_callbacks_split_request.yml'
    )
